	      ; This is a DOS COM file.                      -*-text-*-
	      ; nasm -f bin -o xmsfixer.com xmsfixer.asm

	      CPU 386
	      BITS 16
	      org 100h

%macro print 1; Print $-terminated string
	      mov dx, %1
	      mov ah, 9
	      int 21h
%endmacro

%macro fail 0 ; Exit with code 1
	      mov ax, 0x4C01
	      int 21h
%endmacro

%macro win 0  ; Exit with code 0
	      mov ax, 0x4C00
	      int 21h
%endmacro

              print hello

	      ; XMM installation check requires DOS 3+.
	      ; True DOS version check requires DOS 5+.
	      ; XMS 3.0 was introduced with DOS 6.
	      ; 32-bit registers require 386+.
	      ; W98 itself requires 486+.
	      ;
	      ; How to test if an interrupt vector is set?
	      ; How to check DOS version when doing so requires a version?
	      ; Too much chicken-and-egg... punt it.

	      ; XMS version and DOS version both are returned in a kind of
	      ; packed BCD.  Per XMS docs, 0x0235 would mean version 2.35.
	      ; However, the DOS version is returned in reverse byte order.

	      ; "True" DOS version check.
	      mov ax, 0x3306
	      int 21h
	      cmp bx, 0x0A07 ; means version 7.10
	      jne notW98

	      ; XMM installation check.
	      mov ax, 0x4300
	      int 2Fh
	      cmp al, 0x80
	      jne noxmmfail

              ; Get XMM call address.
	      mov ax, 0x4310
              int 2Fh
              mov word [XMM],   bx
              mov word [XMM+2], es

              ; XMS version check.
              mov ah, 0
              call far [XMM]
	      cmp ah, 2
	      je xmsv2       ; Version 2 can't have the problem.
	      cmp ah, 3
              jne xmsverfail ; Are there versions besides 2 and 3?

	      ; If no 510 MiB block is available, don't do anything.
	      mov ah, 0x88
	      call far [XMM]
	      cmp bl, 0xA0
	      je nofree    ; BL  = A0h if all extended memory is allocated
	      test bl, bl
	      jnz queryfail
	      cmp eax, 0x7F800 ; 510 MiB = 0x7F800 KiB
	      jb noblockfail

	      ; Try to allocate a 510 MiB block, which we *hope* will appear
	      ; at the bottom of extended memory.
	      mov edx, 0x7F800
	      mov ah, 0x89
	      call far [XMM]
	      cmp bl, 0xA1
	      je handlfail
	      cmp ax, 0x0001
	      jne allocfail
	      mov [handle], dx
	      print step1

	      ; Loop until no free XMS.
while:        mov ah, 0x88
	      call far [XMM]
	      cmp bl, 0xA0
	      je wend      ; BL  = A0h if all extended memory is allocated
	      test bl, bl
	      jnz queryfail

	      ; EAX = size of largest free block.
	      ; Allocate it.
	      mov edx, eax
	      mov [didalloc], eax
	      mov ah, 0x89
	      call far [XMM]
	      cmp bl, 0xA1
	      je handlfail
	      cmp ax, 0x0001
	      jne allocfail

	      ; Report successful allocation.
	      print alloca
	      mov eax, [didalloc]
	      call printeax
	      print allocb

	      jmp while

	      ; Free that 510 MiB block.
wend:         mov dx, [handle]
	      mov ah, 0x0A
	      call far [XMM]
	      cmp ax, 0x0001
	      jne freefail

	      ; Done.
              print step3
	      win

; --------------------------------------------------------------------------
; Branches for printing messages and exiting.

notW98:	      print s_notW98
	      win
noxmmfail:    print s_noxmmfail
	      fail
xmsverfail:   push ax
	      print s_xmsverfail
	      pop ax
	      call printax
	      print EOL
	      fail
xmsv2:	      print s_xmsv2
	      win
nofree:	      print s_nofree
	      win
noblockfail:  print s_noblocka
	      call printeax
	      print s_noblockb
	      fail
queryfail:    print s_queryfail
	      call printbl
	      print EOL
	      fail
handlfail:    print s_handlfail
	      fail
allocfail:    print s_allocfail
	      call printbl
	      print EOL
	      fail
freefail:     print s_freefail
	      call printbl
	      print EOL
	      fail

; --------------------------------------------------------------------------
; You'd think there'd be a standard way to print numbers, but no.

              ; Print 32-bit value in eax, preserving eax
printeax:     rol eax, 4
	      call printnybble
              rol eax, 4
	      call printnybble
              rol eax, 4
	      call printnybble
              rol eax, 4
	      call printnybble
              rol eax, 4
	      call printnybble
              rol eax, 4
	      call printnybble
              rol eax, 4
	      call printnybble
              rol eax, 4
	      call printnybble
	      ret

              ; Print 16-bit value in ax, preserving ax
printax:      rol ax, 4
	      call printnybble
              rol ax, 4
	      call printnybble
              rol ax, 4
	      call printnybble
              rol ax, 4
	      call printnybble
	      ret

	      ; Print 8-bit value in bl
printbl:      mov al, bl
              rol al, 4
	      call printnybble
              rol al, 4
	      call printnybble
	      ret

	      ; Print lowest 4 bits of ax, preserving ax
printnybble:  push ax
	      and al, 0Fh
	      cmp al, 0Ah
	      jb lt10
	      add al, 07h
lt10:	      add al, 30h
	      mov dl, al
	      mov ah, 6
	      int 21h
	      pop ax
	      ret

; --------------------------------------------------------------------------

EOL:          db `\r\n$`
hello:        db `W98 XMS fixer 2018-06-24 David Flater (CC BY 4.0)\r\n$`
step1:        db ` Allocated 510 MiB block\r\n$`
alloca:       db ` Allocated XMS block size 0x$`
allocb:       db ` KiB\r\n$`
step3:        db ` Freed 510 MiB block\r\n$`

s_notW98:     db ` DOS version is not W98 - fix not needed\r\n$`
s_noxmmfail:  db ` Fail:  extended memory manager not present\r\n$`
s_xmsverfail: db ` Fail:  unexpected XMS version:  $`
s_xmsv2:      db ` XMS version 2 has 64 MiB limit - fix not needed\r\n$`
s_nofree:     db ` All XMS is allocated already - fix not needed\r\n$`
s_noblocka:   db ` Largest available block = 0x$`
s_noblockb:   db ` KiB - no action taken\r\n$`
s_queryfail:  db ` Fail:  query free XMS returned error code 0x$`
s_handlfail:  db ` Fail:  all extended memory handles are in use\r\n$`
s_allocfail:  db ` Fail:  allocate XMS block returned error code 0x$`
s_freefail:   db ` Fail:  free XMS block returned error code 0x$`

XMM:          dd 0
handle:       dw 0
didalloc:     dd 0
